/*****************************************************************************\
 *  genrand.c,v 1.3 2002/12/09 03:35:32 garlick Exp
 *****************************************************************************
 *  Copyright (C) 2001-2002 The Regents of the University of California.
 *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
 *  Written by Jim Garlick <garlick@llnl.gov>.
 *  UCRL-CODE-2003-006.
 *  
 *  This file is part of Scrub, a program for erasing disks.
 *  For details, see <http://www.llnl.gov/linux/scrub/>.
 *  
 *  Scrub is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *  
 *  Scrub is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with Scrub; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
\*****************************************************************************/

/* 4/99 added reentrant versions of interface for MT version of scrub
 * Todo: verify reentrant versions and make parallel
 */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include "md5.h"
#include "genrand.h"

static char Randpool[16];
static long Randcnt;
static unsigned int Randseed;

#ifdef NEED_RAND_R
static int rand_r(unsigned int *seed)
{
    static int srand_called = 0;
    int retval;

    /* XXX begin crit sec */
    if (!srand_called)
        srand(*seed);
    retval = rand();
    /* XXX end crit sec */

    return retval;
}
#endif

/*
 * Call early and call often on a wide variety of random or semi-
 * random system events to churn the randomness pool.
 * The exact format and length of randevent doesn't matter as long as
 * its contents are at least somewhat unpredictable.
 */
void churnrand_r(char *rp, int rplen, unsigned int *seed)
{
    char randevent[16];
    MD5_CTX md5;
    int i;

    for (i = 0; i < sizeof(randevent); i++)
        randevent[i] = rand_r(seed);

    MD5Init(&md5);
    MD5Update(&md5, rp, rplen);
    MD5Update(&md5, randevent, sizeof(randevent));
    MD5Final(rp, &md5);
}

void churnrand()
{
    churnrand_r(Randpool, sizeof(Randpool), &Randseed);
}

void initrand_r(char *rp, int rplen, unsigned int *seed)
{
    int count;
    /* seed RNG */
    count = rand_r(seed);       /* build up sufficient randomness */
    count = abs(count % 1024);  /* limit the number of iterations */
    while (count--)
        churnrand_r(rp, rplen, seed);
}

void initrand()
{
    Randseed = (unsigned int) time(0);
    initrand_r(Randpool, sizeof(Randpool), &Randseed);
}

void genrand_r(unsigned char *buf, unsigned int buflen, long *rc, char *rp,
               int rplen)
{
    MD5_CTX md5;
    char tmp[16];
    unsigned int n;

    while (buflen != 0) {
        /* Hash the pool with a counter */
        MD5Init(&md5);
        MD5Update(&md5, rp, rplen);
        MD5Update(&md5, (unsigned char *) rc, sizeof(long));
        MD5Final(tmp, &md5);
        (*rc)++;                /* Increment counter */
        /* 
         * Copy 16 bytes or requested amount, whichever is less,
         * to the user's buffer 
         */
        n = (buflen < 16) ? buflen : 16;
        memcpy(buf, tmp, n);
        buf += n;
        buflen -= n;
    }
}

void genrand(unsigned char *buf, unsigned int buflen)
{
    genrand_r(buf, buflen, &Randcnt, Randpool, sizeof(Randpool));
}


#ifdef STAND
int main(int argc, char *argv[])
{
    char buf[24];
    int i, j;

    initrand();
    for (j = 0; j < 20; j++) {
        churnrand();
        genrand(buf, 24);
        for (i = 0; i < 24; i++)
            printf("%x ", buf[i]);
        printf("\n");
    }
    exit(0);
}
#endif

/*
 * vi:tabstop=4 shiftwidth=4 expandtab
 */
